name SER_PORT

;****************************************************************************
;*	   Initial routines, Interrupt service(commands	from host)	    *
;****************************************************************************

$include(Const.h)
$include(Rs232.h)
$include(EEPROM.H)

;usage of register banks
;bank 0		foreground
;bank 1		Serial communication channel interrupt
;bank 2		not used as registers
;bank 3		not used as registers

;==============================================================================

	public		_UARTinit,UARTinit
	public		UARTint
	public		TempBuf
	public		Sec_Op,WakeUp,InitRq,Ignore,RqSerData
	public		SendTempBuf

	public		BadConfig

	extrn	CODE	(Version)
	extrn	CODE	(EErd,EEwr,EEwrEn,EEwrDs)
	extrn	DATA	(SerNumH,SerNumL,EE_Address,EE_DataArea)
	extrn	bit	(EE_wrRQ,EE_rdRQ,EE_DrRQ,EE_DwRQ)

;============================================================================
;			 Serial	channel	services
;============================================================================

?DT?Serial	SEGMENT	IDATA
	RSEG	?DT?Serial	; datov segment Cka

TempBuf:	ds	26
LenTBuf	equ	$-TempBuf


?BA?SFlags	SEGMENT	DATA BITADDRESSABLE
	RSEG	?BA?SFlags	;datov	segment	Cka
	
progst:		ds	1
Sec_op		bit	progst.0
WakeUp		bit	progst.1
InitRq		bit	progst.2
Ignore		bit	progst.3
InvCom		bit	progst.4
DirectRS	bit	progst.5
RqSerData	bit	progst.6
Sync		bit	progst.7

devstat:	ds	1
BadConfig	bit	devstat.0

?PR?SerialInit	segment	code	INBLOCK
	rseg	?PR?SerialInit

;+--------------------------------------------------------------------------+
;+			Initial	of serial channel			    +
;+--------------------------------------------------------------------------+
_UARTinit:
if RS485 = YES
	setb	TxEn		;Input data enabled (RS485 only)
endif
	mov	a,r7		;Save device address
UARTinit:
	push	psw
	mov	psw,#Bank_1	;Select	register bank 1	and init registers
	mov	r0,#TempBuf	;TempBuf address
	mov	r2,#0		;Init byte counter register
	mov	r3,a		;Set device address
	mov	r4,#0		;Init check summ register
	mov	r5,#0		;Init actual received bytes counter
	mov	r6,#2		;Init bytes expected counter
	mov	TL1,#BAUD96	;Baud rate generator
	mov	TH1,#BAUD96	;Baud rate generator - reload value
	mov	SCON,#c_SCON	;Nastav	reim seriovho	portu
				;Init flags
	clr	Ignore		;Clear IGNORE STATE flag
	clr	Sync		;Clear sync flag
	clr	Sec_op		;Clear secondary opcode	flag
	setb	WakeUp		;End of	WakeUp state
	pop	psw		;Restore program status	register
	setb	ES		;Enable	serial port's interrupt
	ret
;+--------------------------------------------------------------------------+

?PR?SerialService	segment	code	INBLOCK
	rseg		?PR?SerialService
	USING	1
;+--------------------------------------------------------------------------+
;+		Comunication channel interrupt service			    +
;+	    Outputs or inputs via temporrary buffer TempBuf		    +
;+--------------------------------------------------------------------------+
;Bank 1	registers:	r0 - Pointer do	TempBuf	by s_out operation
;			r1 - Temporrary
;			r2 - Byte counter by s_out operation
;!!! DON'T CHANGE !!!	r3 - Device address		 !!! DON'T CHANGE !!!
;			r4 - Check sum counter by input/output operation
;			r5 - Input bytes counter
;			r6 - Max input bytes count
;			r7 - Temporrary
UARTint:
	push	psw		;Save program status register
	push	acc		;Save accumulator
	push	b		;Save b
	mov	psw,#Bank_1	;Select	bank 1
	jb	RI,InpChar	;Rec. interrupt? yes->jump
	clr	TI		;Output	of new char
if DEBUG = YES then
	mov	r7,#0		;Delay - FOR DEBUG ONLY	!!!
	djnz	r7,$
endif
	cjne	r2,#0,NextCh	;Any char? yes->jump
	mov	r7,#hold_time	;Hold output AFTER sending
	djnz	r7,$		;Wait !!!
	clr	RI		;Clear possible	invalid	input
if RS485 = YES then
	setb	TxEn		;Input data enabled (RS485 only)
endif
com_ret_org:
	mov	r0,#TempBuf	;Init TempBuf pointer
com_ret_org1:
	clr	Ignore		;Init flag - IGNORE STATE
com_ret_org2:
	mov	r4,#0		;Init input chsumm counter
	mov	r5,#0		;Init input char counter
	mov	r6,#2		;Init max input	char
	clr	Sec_op		;Clear secondary opcode	flag (=primary opcode)
com_ret:
	pop	b		;Restore b
	pop	acc		;Restore accumulator
	pop	psw		;Restore program status	register
	reti			;End of	interrupt service
NextCh:
if RS485 = YES then
	jb	BadConfig,NoTxEn1
	clr	TxEn		;Output	data enabled (RS485 only)
NoTxEn1:
endif
	dec	r2		;Decrement number of output bytes
	mov	a,@r0		;Get output byte
	inc	r0		;Increment pointer
	mov	C,P		;Set parity bit
;	 cpl	 C		 ;Odd parity for send !!!
	mov	TB8,C		;Set parity bit	in tranceiver
	mov	SBUF,a		;Save byte to tranceiver buffer
	sjmp	com_ret		;End of	service
InpChar:			;Interpreter of	input command
	clr	RI
if RS485 = YES then
	jb	BadConfig,AlwaysEnabled
	jnb	TxEn,com_ret	;Received state? no->jump
AlwaysEnabled:
endif
	acall	SBInp		;Get char from serial port
	jc	ParErr		;Parity	error? yes->jump   Check !!! 
;	 jnc	 ParErr		 ;Parity error?	yes->jump   Check !!!
	jnb	Sync,store_byte	;Test synchron.	cycle? no->jump
	cjne	a,#0,sync_cyk1	;Sync cycle continues? no->jump
	jmp	com_ret_org	;Ignore	block
sync_cyk1:
	clr	Sync		;End of	sync cycle
store_byte:
	cjne	r5,#LenTBuf,nextzna ;Temp buffer overflow? no->jump
	sjmp	IgnoreRest	;Ignore	bytes overflowing buffer length
nextzna:
	inc	r5		;Counter of rec. bytes++
	mov	@r0,a		;Store received	byte in	temp buffer
	inc	r0		;Address of next byte
IgnoreRest:
	xch	a,r4		;Compute checksum
	add	a,r4
	xch	a,r4
	jnb	Sec_op,FixComm	;Variable lentgth? (second opcod)? no->jump
	cjne	r5,#5,FixComm	;Was counter of	bytes received ? no->jump
	clr	Sec_op		;Clear Sec_op flag -> Wait for primary opcode
	jz	ComErr		;a=0?	yes->2nd opcode=0 -> jump -> ignore
	cjne	a,#1,cerr1	;a=1?
	ajmp	ComErr		;	yes->2nd opcode=1 -> jump -> ignore
cerr1:
;	cjne	a,#2,cerr2	;a=2?	29.6.97	command	can have 2bytes	only
;	ajmp	ComErr		;	yes->2nd opcode=2 -> jump -> ignore
;cerr2:
	mov	r7,a		;Save acc		
	cjne	a,#LenStr+2+2+1,$+3	;a>LenStr+3 ?
	jnc	ComErr		;	yes->jump (=error)
	mov	a,r7		;a<3,LenStr+3>	length of string
	mov	r6,a		;r6=number of bytes (r6=bytes' counter)
FixComm:
	cjne	r5,#2,Others	;2nd byte received? no->jump
				;Command decoder 1 - setting r6	counter
	cjne	a,#RetString,UseTab ;Secondary opcode request? no->jump
	setb	Sec_op		;	 yes ->	Set secondary opcode flag
UseTab:
	mov	r7,a		;a=15 max. -> max 16 commands in this version
	anl	a,#HiN		;Only high nibble		
	jnz	ComErr		;a>15? yes -> error -> jump
	push	dph		;Save DP register
	push	dpl
	xrl	a,r7		;Only low nibble (max 16 commands!!!)
	mov	DPTR,#Comm_len	;Find length off command for counter
	movc	a,@a+DPTR	;Get command length
	mov	r6,a		;Set command length counter
	pop	dpl		;Restore DP register
	pop	dph
Others:
	djnz	r6,com_ret	;Wait for next byte? yes->jump
	cjne	r4,#0,ChsumE	;r4!=0?	yes->jump (=Checksum error)
	jb	Ignore,com_ret_org ;Ignore? yes->jump
	mov	r0,#TempBuf	;1st byte address -> device address
	cjne	@r0,#0FFh,Decode;Device_address=FFh? ->	no->jump (check	address)
MyAddr:				;Command decoder 2 -> jump table
	push	dph		;Save DP register
	push	dpl
	mov	r0,#TempBuf+1	;Address of opcode
	mov	a,@r0		;Get opcode
	anl	a,#LoN		;Only low nibble -> max	16 commands
	rl	a		;a=2*a
	mov	DPTR,#Comm_br	;Command branch	table
	jmp	@a+DPTR		; -> jump by opcode
Decode:
if RS485 = YES then
	jb	BadConfig,AlwaysMyAdr
	mov	a,@r0
	jz	MyAddr		;Address==0 ? ->Listen!!
	cjne	a,ar3,Decode1	;Compare address with device adr
AlwaysMyAdr:
endif
	ajmp	MyAddr
Decode1:
;	xch	a,r3		;Command was for other device
	mov	r7,#hold_time
	djnz	r7,$		;Dead time before an after returned command
	clr	RI		;Clear possible	invalid	input
	ajmp	com_ret_org
ParErr:				;Set error - Parity error (invalid parity)
;	lcall	set_terr	;Call service routine
;	setb	Ignore		;Invalid byte in command
	mov	r0,#TempBuf	;Init TempBuff pointer
	ajmp	com_ret_org2
ComErr:
	setb	InvCom		;Set invalid command flag
	ajmp	com_ret_org
ChsumE:
;	acall	set_chsum	;Checksumm error indicator
	ajmp	com_ret_org
dek_ret:
	pop	dpl		;Restore DP register
	pop	dph
	ajmp	com_ret_org	;

dek_ret1:
	pop	dpl		;Restore DP register
	pop	dph
	ajmp	com_ret_org1

;-----<	Commands length	tale >-----------------------------------------------
;NOTE: Length includes opcode, but NOT check summ !!!
Comm_len:
	db	2		;0 Enter sync cycle		
	db	2		;1 Device initial		
	db	2		;2 Return version		
	db	2		;3 Return status		
	db	4		;4 Read	data word from EEPROM	
	db	6		;5 Write data word to EEPROM	
	db	4		;6 Read	DALLAS CODE from EEPROM		      
	db     12		;7 Write DALLAS	CODE to	EEPROM
	db	4		;8
	db	8		;9
	db	2		;A
	db	2		;B n/a
	db	2		;C n/a
;-----<	Answers	>------------------------------------------------------------
RetByte	equ	$-Comm_len	;D Returned byte
	db	3
;----------------------------------------------------------------------------
RetWord	equ	$-Comm_len	;E Returned word 
	db	4
;----------------------------------------------------------------------------
RetString equ	$-Comm_len	;F Data	with variable length
	db	4		;Write up to LenStr bytes to the scratchpad mem.
;============================================================================
;
;command branch	table
comm_br:
	AJMP	opc0		;0 Enter sync cycle		
	AJMP	opc1		;1 Device initial			  
	AJMP	opc2		;2 Return version		
	AJMP	opc3		;3 Return status		
	AJMP	opc4		;4 Read	data word from EEPROM	
	AJMP	opc5		;5 Write data word to EEPROM	
	AJMP	opc6		;6 Read	DALLAS CODE from EEPROM	
	AJMP	opc7		;7 Write DALLAS	CODE to	EEPROM	
	AJMP	opc8		;8 
	AJMP	opc9		;9
	AJMP	opcA		;A 
	AJMP	opcB		;B
	AJMP	dek_ret		;C 
;-----<	Answers	>------------------------------------------------------------
	AJMP	dek_ret		;D Returned byte (not used as command) 
	AJMP	dek_ret		;E Returned word (not used as command)
	AJMP	opcF		;F Data	with variable length

;+==========================================================================+
;+			     Interface init				    +
;+		clears input buffer and	error indicators		    +
;+==========================================================================+
opc0:				;Background command service - must be fast !
	setb	Sync		;Set synchronization cycle
opc0r:	ajmp	dek_ret

;+==========================================================================+
;+		 Device	init (clears output buffer,...)			    +
;+==========================================================================+
opc1:
	jnb	WakeUp,End1	;Wake up state?	yes->jump
	setb	InitRq		;Break to system initialization
End1:	ajmp	dek_ret		;End

;+==========================================================================+
;+			  Send software	version				    +
;+==========================================================================+
opc2:
	jnb	WakeUp,End2	;Wake up state?	yes->jump
if RS485 = YES then
	jb	BadConfig,NoTxEn2
	clr	TxEn		;Transmiter enabled
NoTxEn2:
endif
	mov	r0,#TempBuf	;Output	buffer address
;	 mov	 @r0,ar3	 ;My address
	mov	@r0,#MaAddr	;Master	address
	inc	r0		;Next byte address
	mov	@r0,#RetString	;Op Code
	inc	r0		;TempBuf+2 without change
	inc	r0		;TempBuf+3 without change
	inc	r0		;
	mov	r1,ar0		;save for future use
	inc	r0		;TempBuf+5 
	mov	DPTR,#Version	;Address of version string
	acall	CODEtoRAM	;Move version string to	temporary buffer
	inc	r2		;Add two bytes	( Number of bytes )
	inc	r2		;		( Checksum byte	 )
	mov	a,r2
	mov	@r1,a		;Count of bytes	incl. sec. opcode and chsumm
	dec	r0		;Step on 00h
	dec	r0		;Step on space before 00h
	mov	r6,SerNumH
	mov	r7,SerNumL
	mov	r5,#10		;Convert BIN number to DEC radix
lopc2_1:
	acall	XDIV		;r6r7/r5->r6r7 int part	of result, r4 remainder
	mov	a,#'0'
	add	a,r4		;Value convert to ASCII	'0'..'9' 
	mov	@r0,a		;Save ASCII number to output string 
	dec	r0		;address of next number	
	mov	a,r6
	orl	a,r7		;r6r7!=0 ?
	jnz	lopc2_1		;	   yes->jump

	inc	r2
	inc	r2
	inc	r2
	mov	r0,#TempBuf
	acall	StartSend
End2:	ajmp	dek_ret1

;+==========================================================================+
;+			   Return device status				    +
;+==========================================================================+
opc3:
	jnb	WakeUp,End3	;Wake up state?	yes->jump
if RS485 = YES then
	jb	BadConfig,NoTxEn3
	clr	TxEn		;Transmiter enabled
NoTxEn3:
endif
	mov	r0,#TempBuf	;Output	buffer address
;	 mov	 @r0,ar3	 ;My address
	mov	@r0,#MaAddr	;Master	address
	inc	r0		;Next byte address
	mov	@r0,#RetByte	;Op code
	inc	r0		;Next byte address
	inc	r0		;TempBuf+2 without change
	inc	r0		;TempBuf+3 without change
	mov	a,devstat	;Device	status to accumulator
AnswerD:
	mov	@r0,a		;Device	status to buffer
	mov	r2,#5		;Number	of bytes
	mov	r0,#TempBuf	;Source	address
	acall	StartSend	;Compute checksum and start send
End3:	ajmp	dek_ret1	;End with answer

;+==========================================================================+
;+			   Write data to EEPROM				    +
;+==========================================================================+
opc5:
	jnb	WakeUp,End4	;Wake up state?	yes->jump
	mov	r0,#TempBuf
	mov	@r0,#MaAddr	;TempBuf+0 <- Master address
	inc	r0		;
	mov	@r0,#RetWord	;TempBuf+1 <- Op code
	inc	r0		;TempBuf+2
	inc	r0		;TempBuf+3
	mov	a,@r0		;a  <- (TempBuf+3)
	inc	r0		;TempBuf+4
	mov	EE_Address,a	;Prepare for use
	mov	r2,a		;r2 <- EEPROM word address
	mov	r1,#TempBuf+10	;Move data here
	call	EErd
	mov	r1,#TempBuf+10
	mov	a,@r1
	xrl	a,@r0		;EEPROM(r2) = data to write ?
	jnz	opc5_W		;			     no	-> write
	inc	r1
	mov	a,@r1
	inc	r0
	xrl	a,@r0		;EEPROM(r2) = data to write ?
	jz	opc4		;			     yes->don't	write
opc5_W:
	setb	EE_wrRQ		;Request for write data	to EEPROM
	mov	EE_DataArea,#TempBuf+4
	ajmp	dek_ret		;End without answer now. Main program start
				;send data automaticaly	after writing

;+==========================================================================+
;+			   Read	data from EEPROM			    +
;+==========================================================================+
opc4:
	jnb	WakeUp,End4	;Wake up state?	yes->jump
if RS485 = YES then
	jb	BadConfig,NoTxEn4
	clr	TxEn		;Transmiter enabled
NoTxEn4:
endif
	setb	EE_rdRQ		;Request for read data from EEPROM
	mov	r0,#TempBuf	;Output	buffer address (=Prepare to Send byte)
	mov	@r0,#MaAddr	;Master	address
	inc	r0		;Next byte address
	mov	@r0,#RetWord	;Op code
	inc	r0		;Next byte address
	inc	r0		;TempBuf+2 without change
	mov	EE_Address,@r0
	mov	EE_DataArea,#TempBuf+4
End4:
	ajmp	dek_ret		;End without answer now. Main program start
				;send data automaticaly	after reading

;+==========================================================================+
;+			Write DALLAS CODE to EEPROM			    +
;+==========================================================================+
opc7:
	jnb	WakeUp,End7	;Wake up state?	yes->jump
	mov	r0,#TempBuf
	mov	@r0,#MaAddr	;Master	address
	inc	r0		;Next byte address
	mov	@r0,#RetString	;Op code
	inc	r0		;Ignore	Hi byte	of DALLAS CODE number 
	inc	r0		;which DALLAS code
	mov	EE_Address,@r0
	inc	r0
	mov	EE_DataArea,r0
	mov	a,EE_Address
	cjne	a,#c_maxDALLAS+1,$+3	;Test on max. number DALLAS CODEs in EEPROM
	jnc	MaxNumberReached
	setb	EE_DwRQ		;Request for read DALLAS CODE from EEPROM
End7:
	ajmp	dek_ret		;without answer
;+==========================================================================+
;+			Read DALLAS CODE from EEPROM			    +
;+==========================================================================+
opc6:
	jnb	WakeUp,End6	;Wake up state?	yes->jump (not answer!!)
if RS485 = YES then
	jb	BadConfig,NoTxEn6
	clr	TxEn		;Transmiter enabled
NoTxEn6:
endif
	mov	r0,#TempBuf
	mov	@r0,#MaAddr	;Master	address
	inc	r0		;Next byte address
	mov	@r0,#RetString	;Op code
	inc	r0		;Ignore	Hi byte	of DALLAS CODE number 
	inc	r0		;which DALLAS code
	mov	EE_Address,@r0
	inc	r0
	mov	EE_DataArea,#TempBuf+5
	mov	a,EE_Address
	cjne	a,#c_maxDALLAS+1,$+3	;Test on max. number DALLAS CODEs in EEPROM
	jnc	MaxNumberReached
	mov	@r0,#0Ah
	setb	EE_DrRQ		;Request for read DALLAS CODE from EEPROM
End6:
	ajmp	dek_ret		;End without answer now. Main program start
				;send data automaticaly	after reading
MaxNumberReached:
	mov	@r0,#02h
	mov	r2,#5
	call	SendTempBuf
	ajmp	dek_ret1	;End with answer

;+==========================================================================+
;+									    +
;+==========================================================================+
opc8:
	jb	WakeUp,End9   ;Wake up state? no->jump
	
	ajmp	dek_ret

;+==========================================================================+
;+									    +
;+==========================================================================+
opc9:
	jnb	WakeUp,End9	;Wake up state?	yes->jump
				;not implemented yet
End9:	ajmp	dek_ret		;End without answer

;+==========================================================================+
;+									    +
;+==========================================================================+
opcA:
	jnb	WakeUp,opc0r1
opc0r1:
				;not implemented yet
	ajmp	dek_ret		;End without answer

;+==========================================================================+
;+									    +
;+==========================================================================+
opcB:
	jnb	WakeUp,opc0r1
dek_ret_endc:
	ajmp	dek_ret		;End without answer
if RS485 = YES then
	jb	BadConfig,NoTxEnB
	clr	TxEn		;Transmiter enabled
NoTxEnB:
endif
				;not implemented yet
	ajmp	dek_ret1	;End with answer
;+==========================================================================+
;+									    +
;+==========================================================================+
opcD:
	jnb	WakeUp,opc0r1
	ajmp	dek_ret		;End without answer

;+==========================================================================+
;+									    +
;+==========================================================================+
opcE:
	jnb	WakeUp,opc0r1
	ajmp	dek_ret		;End without answer

;+==========================================================================+
;+		Write up to LenStr bytes to the	scratchpad memory		    +
;+==========================================================================+
opcF:
	jnb	WakeUp,opc0r1
	ajmp	dek_ret		;End without answer


;+--------------------------------------------------------------------------+
;+	 Computes check	summ and starts	output of blok to s_channel	    +
;+--------------------------------------------------------------------------+
;r0  =	pointer	to first byte
;r2  =	info bytes count
;r7  =	temp, delay

SendTempBuf:
	mov	r0,#TempBuf
StartSend:

if RS485 = YES then
	jb	BadConfig,NoTxEnStartSend
	clr	TxEn		;Transmiter enabled
NoTxEnStartSend:
	mov	r7,#turn_time
	djnz	r7,$		;Delay before sending
endif
	mov	a,r0
	push	acc
	mov	a,r2
	push	acc

	mov	a,#0		;Clear check summ acumulator
chsumcyk:
	add	a,@r0		;Compute check sum
	inc	r0		;Address of next byte
	djnz	r2,chsumcyk	;Next byte? yes->jump
	cpl	a		;First complement
	add	a,#1		;Second	complement
	mov	@r0,a		;Save check summ
	pop	acc		;Restore r2 reg
	mov	r2,a
	pop	acc		;Restore r0 reg
	mov	r0,a
	mov	a,@r0		;Get byte for send
	clr	TI
	add	a,#0
	mov	C,P
;	 cpl	 C		 ;Odd parity for send !!!		28.1.98
	mov	TB8,C		;Set parity bit
	mov	SBUF,a		;Start of transfer	
	inc	r0		;Address of next byte
	ret	

;+--------------------------------------------------------------------------+
;	Input byte from	Serial port
;+--------------------------------------------------------------------------+
;Input:	-
;Output:acc - received byte
;	C=0 - parity OK.
;	C=1 - parity error
;Change:acc, C 

SBInpX:				;SBInpX	entry is used for debug	only
	jnb	RI,$		;Wait for byte
	clr	RI		;Clear flag
SBInp:	mov	a,SBUF		;Get received byte
	mov	C,P		;Parity	check
	jnb	RB8,No_Change	;
	cpl	C
No_Change:
	ret

;==============================================================================
;transfers string from DPTR in CODE memory to @r0 - string ends	0
;On return r2 =	byte transfered	incl. 0
CODEtoRAM:
	mov	r2,#0
trfis12:
	clr	a
	movc	a,@a+DPTR
	mov	@r0,a
	inc	DPTR
	inc	r0
	inc	r2
	jnz	trfis12
	ret

;+--------------------------------------------------------------------------+
;	Unsigned division r6r7/r5
;+--------------------------------------------------------------------------+
;Input:	r5  - divider
;	r6  - high byte	of divisor
;	r7  - low byte of divisor
;Output:r4  - remainder
;	r5  - divider (not changed)
;	r6  - high byte	of result
;	r7  - low byte of result
;Change:acc,b,r4,r5,r6,r7
 
XDIV:
	cjne	r6,#0,lUD0
	mov	a,r7		;r6=0 ->r7/r5
	mov	b,r5
	div	ab
	mov	r7,a
	mov	r4,b
	ret
lUD0:
	mov	a,r5
	mov	r4,a
	mov	b,a
	mov	a,r6
	div	ab
	jb	OV,lUD9
	mov	r6,a
	mov	r5,b
	mov	b,#8
lUD1:
	mov	a,r7
	add	a,r7
	mov	r7,a
	mov	a,r5
	rlc	a
	mov	r5,a
	jc	lUD2
	subb	a,r4
	jnc	lUD3
	djnz	b,lUD1
	sjmp	lUD8
lUD2:
	clr	C
	subb	a,r4
lUD3:
	mov	r5,a
	inc	r7
	djnz	b,lUD1
lUD8:
	mov	a,r4
	xch	a,r5
	mov	r4,a
lUD9:
	ret

;==============================================================================
	end

