
$include(CONST.H)

;*****************************************************************************
;**********************< TOUCH MEMORY INTERFACE	>*****************************

	public	TMtest,TMrst,TMrd,TMwr,CalcCRC,EntCRC

	extrn	bit	(TMpre)

TouchM	segment	code	INBLOCK
	rseg	TouchM

;+--------------------------------------------------------------------------+
;+------	Touch memory - Presence	test				----+
;+--------------------------------------------------------------------------+
;Input:
;Output:CrdPre=1 if TM is present
;	      =0 if TM is absent
TMtest:
	clr	TMpre		;Clear present flag
	acall	TMrst		;Test for presence TM
	jnc	TMtst_9		;The TM	is absent ->jump (Error)
	setb	TMpre		;The TM	is present.
TMtst_9:
	ret

;+--------------------------------------------------------------------------+
;+------	Touch memory reset					----+
;+--------------------------------------------------------------------------+
TMrst:
	clr	EA
	setb	TMdata		;Enable	TM data	input
	mov	r7,#40/2	;Load recovery time counter
	djnz	r7,$		;Wait
	clr	TMdata		;Start the reset pulse
	mov	r7,#480/2	;Load width of the reset pulse
	djnz	r7,$		;Wait -	generate the reset pulse
	setb	TMdata		;Release TM data line -	enable detect touch memory
	mov	r7,#12/4	;Set time interval for recovery	TM data	line
TMrecovery:
	jb	TMdata,TMcheckPresentPulse
	djnz	r7,TMrecovery	;Wait
	sjmp	TMabsent
TMcheckPresentPulse:
	mov	r7,#80/4	;Max. delay for	detection of presence pulse
TMchk0:
	jnb	TMdata,TMchk1;Catch presence pulse
	djnz	r7,TMchk0	;Max. delay for	detect present pulse
	sjmp	TMabsent
TMchk1:				;Present pulse was detected
	mov	r7,#55/4
TMchk2:
	jb	TMdata,TMabsent	;Test state of data line (presence pulse)
	djnz	r7,TMchk2
				;Min width of a	TM present pulse was reached
	mov	r7,#230/4	;Max. width of presence	pulse (~240us)
TMchk3:
	jb	TMdata,TMchk4	;Check width of	presence pulse
	djnz	r7,TMchk3
	sjmp	TMabsent
 
	mov	r7,#40/4
TMchk4:
	jb	TMdata,TMpresent;Check end of presence pulse
	djnz	r7,TMchk4	;Max width of a	TM present pulse
TMabsent:
	clr	C		;TM wasn't detected
	setb	EA
	ret
TMpresent:
	setb	C		;TM was	detected succefully
	setb	EA
	ret
	
;+--------------------------------------------------------------------------+
;+------	Touch memory write/read					----+
;+--------------------------------------------------------------------------+
TMrd:	mov	a,#-1
TMwr:
	setb	TMdata		;Enable	TM data	input
	mov	r7,#10/2	;Recovery time ~10us
	djnz	r7,$		;Wait
	mov	r6,#8		;8 bits	will be	write/read
Bit_0:	
	rrc	a		;Move one bit in carry
	clr	EA
	clr	TMdata		;Start the time	slot
	mov	r7,#(8-1)/2	;Width of start	pulse
	djnz	r7,$		;
	mov	TMdata,C	;Send out the data bit
	mov	r7,#(18-8-2)/2	;Delay for read
	djnz	r7,$		;
	mov	C,TMdata	;Read data bit
	mov	r7,#(80-18)/2	;Delay until the end
	djnz	r7,$		;	of the time slot
	setb	TMdata		;Terminate time	slot
	setb	EA
	mov	r7,#2		;Recovery time ~10us
	djnz	r7,$		;
	djnz	r6,Bit_0
	rrc	a
	ret

;+--------------------------------------------------------------------------+
;+------	Touch memory CRC calculation				----+
;+--------------------------------------------------------------------------+
;Input:	r0 - address of	DALLAS code
;Output:r5 - CRC
;Change:a,r5,r6,r7

CalcCRC:
	mov	r6,#7
EntCRC:
	mov	r5,#0
CCRC:	mov	a,@r0
	inc	r0

	push	acc
	push	acc
	mov	r7,#8
CRC_Loop:
	xrl	a,r5
	rrc	a
	mov	a,r5
	jnc	ZERO
	xrl	a,#18h
ZERO:
	rrc	a
	mov	r5,a
	pop	acc
	rr	a
	push	acc
	djnz	r7,CRC_Loop
	pop	acc
	pop	acc
	djnz	r6,CCRC
	ret


	end			;
